/************************************************************************/
/*																		*/
/*                     Bootloader Programmer                            */
/*      Linux-portation by Sascha Biedermann, sascha@coder-area.de      */
/*         Originally written by Peter Dannegger, danni@specs.de        */
/*																		*/
/************************************************************************/

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "readargs.h"

#define DEVICE "/dev/ttyS1"


#define FALSE 0
#define TRUE 1

#define BAUD_DEF 16
#define BAUD_CNT 17

#define KEY_ENTER		0x0D
#define COMMAND_CONTINUE	'c'
#define COMMAND_FAIL		'f'
#define COMMAND_DONE		'a'

#define TIMEOUT		8  // 1 = 100ms

unsigned int baudrate = BAUD_DEF;
char device[] = DEVICE;
speed_t baud_const[BAUD_CNT] = {
								B50,B75,B110,B134,B150,B200,B300,B600,B1200,
            		    	    B1800,B2400,B4800,B9600,B19200,B38400,B57600,
                        	    B115200
	                       };

unsigned long baud_value[BAUD_CNT] = {
					50,75,110,134,150,200,300,600,1200,
					1800,2400,4800,9600,19200,38400,57600,
					115200
				     };


struct termios oldtio;

int _argc;
char **_argv;
int fd;

void program( FILE *flash, FILE *eeprom );
int readhex( FILE *fp, unsigned int *addr, unsigned char *data);
int sscanhex( unsigned char *str, unsigned int *hexout, int n );
void com_putc(unsigned char c);
void com_puts(char *text);
void connect( void );
char com_getc(int timeout);
void com_open(char device[], speed_t baud);
void com_close(void);
void helptxt(void);

int main(int argc, char *argv[])
{
	char arg[100];
	
	FILE *flash=NULL, *eeprom=NULL;
	
	_argc = argc;
	_argv = argv;
	
	
	// Hilfe anzeigen ?
	if (readargs( ABOOL, 'h', &arg )) helptxt();
	
	// Was wollen wir machen?
	if ( !readargs( ABOOL, 'p', &arg)) 
	{
		printf("Nothing to do!");
		helptxt();
	}
	// Baudrate einlesen
	unsigned long tmp_baud;	
	if (readargs( ALONG, 'b', &tmp_baud))
	{		
		unsigned int i, found=FALSE;
		// Prfen, ob die Baudrate gltig ist
		for (i=0;i<BAUD_CNT;i++)
		{
			if (baud_value[i] == tmp_baud)
			{	
				baudrate = i;
				found = TRUE;
				break;
			}
		}
		if (found==FALSE)
		{
			printf("Ungltige Baudrate!");
			helptxt();
		}
	}
	// Device einlesen
	if (readargs( ASTRING, 'd', &arg ) )
	{
		strcpy(device, arg);
	}

	unsigned char file_av=FALSE;
	// Flash-Datei einlesen
	if ( readargs(ASTRING,'f',&arg) )
	{
		file_av=TRUE;
		flash = fopen(arg,"r");
		if (flash == NULL)
		{
			printf("Could not open File \"%s\"!\n",arg);
			exit(-1);
		}		
	}

	// EEPROM-Datei einlesen
	/*if ( readargs(ASTRING,'e',&arg) )
	{
		file_av=TRUE;
		eeprom = fopen(arg,"r");
		if (flash == NULL)
		{
			printf("Could not open File \"%s\"!\n",arg);
			exit(-1);
		}		
	}*/
	
	// wurde berhacom_put eine Datei angegeben?
	if (!file_av)
	{
		printf("No Flash and EEPROM File!");
		helptxt();
	}

	// Sooo... jetzt habe wir alles, auf zum Programmieren...

	com_open(device,baud_const[baudrate]);	
	if (fd < 0)
	{
		perror(device);
		exit(-1);
	}
	
	connect();
	
	
	program(flash,eeprom);
	
	com_close();
	exit(0);
}

void program( FILE *flash, FILE *eeprom )
{
	FILE *fp;
	unsigned int firstaddr = 65536;
	unsigned int lastaddr = 0;
	unsigned int addr;
	int i;
	unsigned char *data;
	unsigned char s[255];
	unsigned char d1,c;
	fp = flash;
	// Erstmal ein wenig Speicher holen
	data = (unsigned char*) malloc( 65536 );
	if( data == NULL )
	{
		printf("Memory allocation error!" );
		return;
	}
	memset(data, 0xFF, 65536);
	while( (i = readhex( fp, &addr, s )) >= 0 )
	{
		if( i )
		{
			memcpy( data + addr, s, i );
			if( firstaddr > addr ) firstaddr = addr;
			addr += i;
			if( lastaddr < addr-1 ) lastaddr = addr-1;
			addr++;
		}
	}
	printf( "Programming: %04X - %04X\n", firstaddr, firstaddr);
  
	com_puts( "DM8" );
  
	tcdrain(fd);
	
	if( com_getc (TIMEOUT) != COMMAND_DONE )
	{
		printf( "Programmer Error 1\n" );
	}
	else
	{
		sprintf( s, "FPR%04X", firstaddr );
		com_puts( s );
		for( addr = firstaddr;; addr++ )
		{
			d1 = data[addr];
			
			//com_putc( d1 );
			write(fd,&d1,1);
			
			//if( d1 == 0xA5 ) com_putc( 0xA5 );			// send A5 two times
			if( d1 == 0xA5 ) write(fd,&d1,1);			// send A5 two times
			
			if( (addr & 0xF) == 0xF )
			{
				tcdrain(fd);
				printf( "\033[AProgramming: %04X - %04X\n", firstaddr, addr + 1);				
				c = com_getc (TIMEOUT);
				if( c != COMMAND_CONTINUE )
				{
					printf( "\nProgrammer Error 2\n");
					break;
				}
			}
			if( addr == lastaddr )
			{
				com_putc( 0xA5);
				com_putc( 0x00);
				tcdrain(fd);
				
				printf( "\033[AProgramming: %04X - %04X\n", firstaddr, addr);	
				
				if( com_getc (TIMEOUT) == COMMAND_DONE )
				{
					printf( "Programming successful" );
				}
				else
				{
					printf( "Programmer Error 3\n" );
				}
				break;
			}
		}
	}
	free( data );
	printf("\n");
}

void connect( void )
{
	char WAITSTRING[] = { '|', '/', '-', '\\' };
	unsigned int i = 1;
	printf("\n");
	while (1)
	{
		printf("\033[A%s at %ld Baud: %c\n", device, baud_value[baudrate],WAITSTRING[++i&3] );
		com_puts( "\370\360\340\200\374" );
		
		if( com_getc( 0 ) == COMMAND_DONE )
		{
			printf("Connected\n");
			com_getc ( 2 );
			tcflush(fd,TCIOFLUSH);
			return;
		}
	}
}



void com_open(char device[], speed_t baud)
{	
	struct termios newtio;
	
	// Erstmal das Device ffnen
	fd = open(device, O_RDWR | O_NOCTTY | O_NDELAY);
	if (fd < 0)
	{
		return;
	}
	
	// Jetzt die Einstellungen Speichern
	tcgetattr(fd,&oldtio);
	
	// Den Neuen Struct mit 0 initzialisieren
	//memset(&newtio, 0x00 , sizeof(newtio));
	
	// Flags setzen
	newtio.c_cflag = CS8 | CLOCAL | CREAD;
	newtio.c_iflag = IGNPAR | IGNBRK;
	newtio.c_oflag = 0;	
	newtio.c_lflag = 0;
	
	//cfmakeraw(&newtio);
	
	// Timeout in 100ms
	newtio.c_cc[VTIME] = 0; 
	// 1 Zeichen lesen
	newtio.c_cc[VMIN] = 0; 
	
	// Baudrate einstellen
	cfsetispeed(&newtio, baud);
	cfsetospeed(&newtio, baud);
	
	// Sende- und Empfangspuffer leeren
	tcflush(fd, TCIOFLUSH);
	
	// die neuen Port-Einstellungen setzen
	tcsetattr(fd,TCSANOW,&newtio);	
	
	return;
}

void com_close(void)
{
	// die alten Einstellung wiederherstellen
	tcsetattr(fd,TCSANOW,&oldtio);
	
	// und das Device schliessen
	close(fd);
}


void helptxt(void)
{
    
	unsigned int i=0;
	printf(	"\n"
			"Usage: %s [-d<device>] [-b<baudrate>] [-p|-r|-v] [[-f<file>]|[-e<file>]]\n"
			"Example: %s -d/dev/ttyS0 -b9600 -p -fmain.hex\n\n"
			"\t-h\t\t shows this help\n"
			"\t-d<device>\t specify the serial device\n"
			"\t\t\t default: /dev/ttyS1\n"
			"\t-b<baudrate>\t specify the baudrate\n"
			"\t\t\t default: 115200\n"
			"\t\t\t valid values are:\n\t\t\t\t",_argv[0],_argv[0]);
	for (i=0; i<BAUD_CNT;i++)
	{
		printf("%d ",(int) baud_value[i]);
		if ( (i%5==0) && (i!=0) ) printf("\n\t\t\t\t");
	}
		
	printf( "\n"
			"\t-f<file>\t specify the flash-file (ihex)\n"
			"\t-e<file>\t specify the eeprom-file (ihex) - not supported yet\n"
			"\t-p\t\t program the device\n"
			"\t-v\t\t verify the device - not supported yet\n"
			"\t-r\t\t read the device - not supported yet\n"
			);
			
	printf("\nLinux-portation by Sascha Biedermann, sascha@coder-area.de\n"
	       "Originally written by Peter Dannegger, danni@specs.de\n");
	  exit(255);
}


void com_putc(unsigned char c)
{  
  tcdrain(fd);
  write(fd,&c,1);  
}

void com_puts(char *text)
{
  while( *text )
    com_putc( *text++ );
  com_putc( KEY_ENTER );		
}
char com_getc(int timeout)
{
	char c;
	clock_t t = clock();

	c = 0;
	while ( !c )
	{
		read(fd,&c,1);
		if( ((clock() - t )/CLOCKS_PER_SEC) >= timeout ) break;
	}
	if (!c) return -1;
	return c;
}

int sscanhex( unsigned char *str, unsigned int *hexout, int n )
{
  unsigned int hex = 0, x = 0;
  for(; n; n--){
    x = *str;
    if( x >= 'a' )
      x += 10 - 'a';
    else if( x >= 'A' )
      x += 10 - 'A';
    else
      x -= '0';
    if( x >= 16 )
      break;
    hex = hex * 16 + x;
    str++;
  }
  *hexout = hex;
  return n;					// 0 if all digits read
}


int readhex( FILE *fp, unsigned int *addr, unsigned char *data)
{
  /* Return value: 1..255	number of bytes
			0	end or segment record
		       -1	file end
		       -2	error or no HEX-File */
  char hexline[524];				// intel hex: max 255 byte
  char * hp = hexline;
  unsigned int byte;
  int i;
  unsigned int num;

  if( fgets( hexline, 524, fp ) == NULL )
    return -1;					// end of file
  if( *hp++ != ':' )
    return -2;                                  // no hex record
  if( sscanhex( hp, &num, 2 ))
    return -2;					// no hex number
  hp += 2;
  if( sscanhex( hp, addr, 4 ))
    return -2;
  hp += 4;
  if( sscanhex( hp, &byte, 2 ))
    return -2;
  if( byte != 0 )				// end or segment record
    return 0;
  for( i = num; i--; ){
    hp += 2;
    if( sscanhex( hp, &byte, 2 ))
      return -2;
    *data++ = byte;
  }
  return num;
}
